#!/bin/sh
# Copyright (c) 2000-2015 Synology Inc. All rights reserved.

CHAIN_NAME_NAT="RESOLVE_SRM_PREROUTING"
DNS_PID_FILE="/var/run/resolve_srm_dns.pid"

BIN_DNSMASQ="/usr/syno/sbin/dnsmasq"
BIN_IPTABLES_V4="/sbin/iptables"

#
# Write a message to stderr.
#
# @param[in] message
#       The log message.
#
Log()
{
	local message="${1}"
	local baseName

	baseName="$(basename "${0}")"

	if [ $# -lt 1 ]; then
		>&2 echo "${baseName} $(date +%F-%T): Log failed: invalid argument"
		logger -p err "${baseName}: Log failed: invalid argument"
		return 1
	fi

	>&2 echo -e "${baseName} $(date +%F-%T): ${message}"
	logger -p err "${baseName}: ${message}"
}

#
# Stop a DNS server.
#
# @see stop_dns_server
#
stop_dns_server ()
{
	# stop the DNS server which is used to resolve router.synology.com
	# on Synology router
	if [ -f ${DNS_PID_FILE} ]; then
		local dns_pid

		dns_pid="$(cat ${DNS_PID_FILE})"

		if [ -d "/proc/${dns_pid}" ]; then
			kill -KILL "${dns_pid}"
		fi
	fi

	iptables -t nat -F ${CHAIN_NAME_NAT}
}

#
# Start a DNS server.
#
# @see stop_dns_server
#
start_dns_server ()
{
	# the dns_hex_string represents:
	#	06 characters: router
	#	08 characters: synology
	#	03 characters: com
	# which is used to match the "router.synology.com" of a DNS query
	local dns_hex_string="|06726f757465720873796e6f6c6f677903636f6d|"
	local new_context="${1}"
	local service_port=5440

	if [ "$#" -lt 1 ]; then
		Log "Error: start_dns_server failed: invalid argument"
		return 1
	fi

	case "${new_context}" in
		"bridge"|"wds")
			local is_enabled

			is_enabled="$(get_key_value /etc/synoinfo.conf enable_bridge_DNS)"

			if [ "xyes" != "x${is_enabled}" ]; then
				return 0
			fi
			;;
		"repeater")
			# do nothing in repeater mode
			return 0
			;;
		*)
			# router mode
			;;
	esac

	Log "update resolving router.synology.com"

	# redirect DNS requests to the DNS server for resolving router.synology.com
	if ! "${BIN_IPTABLES_V4}" -t nat -A "${CHAIN_NAME_NAT}" -p udp --dport 53 \
			-m string --hex-string "${dns_hex_string}" --algo bm \
			-j REDIRECT --to-ports "${service_port}"; then
		Log "Error: failed to redirect UDP port 53 traffic"
		stop_dns_server
		return 1
	fi

	if ! "${BIN_IPTABLES_V4}" -t nat -A "${CHAIN_NAME_NAT}" -p tcp --dport 53 \
			-m string --hex-string "${dns_hex_string}" --algo bm \
			-j REDIRECT --to-ports "${service_port}"; then
		Log "Error: failed to redirect TCP port 53 traffic"
		stop_dns_server
		return 1
	fi

	# start a DNS server to resolve router.synology.com.
	# the resolved IP address "169.254.1.10" can be any address,
	# since it will be modified by resolvsrmd later
	${BIN_DNSMASQ} --user=root --port=${service_port} --pid-file=${DNS_PID_FILE} \
			"--address=/router.synology.com/169.254.1.10" "--address=/router.synology.com/fe80::1"

	return 0
}

#
# Restart the "resolving router.synology.com" service.
#
restart_service()
{
	local szf_script_lock="/tmp/resolve_srm_dns.sh.lock"
	local new_context="${1}"

	if [ "$#" -lt 1 ]; then
		Log "Error: restart_service failed: invalid argument"
		return 1
	fi

	(
		if flock -n -x 8; then
			while true; do
				stop_dns_server
				if start_dns_server "${new_context}"; then
					break
				fi

				Log "failed to start service of resolving router.synology.com"
				sleep 3
			done

			flock -u 8
		fi
	)8> ${szf_script_lock}
}

case $1 in
	--sdk-mod-ver)
		# Print SDK support version
		echo "1.0";
	;;
	--name)
		# Print package name
		echo "A DNS server to resolve router.synology.com."
	;;
	--pkg-ver)
		# Print package version
		echo "1.0";
	;;
	--vendor)
		# Print package vendor
		echo "Synology";
	;;
	--restart)
		# restart the service in background to prevent blocking callers
		new_context="$(synonettool --topology --get)"
		Log "obtain current context [${new_context}] from tool"
		restart_service "${new_context}" &
	;;
	--post)
		old_context="$(/bin/jq -r .old.context "${ARGS_FILE_PATH}")"
		new_context="$(/bin/jq -r .new.context "${ARGS_FILE_PATH}")"

		# restart the service in background to prevent blocking callers
		if [ "${old_context}" != "${new_context}" ]; then
			restart_service "${new_context}" &
		fi
	;;
	*)
		echo "Usage: $0 --sdk-mod-ver|--name|--pkg-ver|--vendor|--pre|--post"
	;;
esac
